from .json_stream import json_stream


def test_json_stream():
    parts = [
        '{"oh no this json',
        '":"is broken"}\n',
        '{"this one":"is not"}\n',
        'and this wont be processed',
    ]
    chunks = [c for c in json_stream(parts)]
    assert len(chunks) == 2


def test_json_stream_fail():
    parts = [
        '{"oh no this json', '":"is broken"}\n',
        'wtf is this?\n'
        '{"this one":"is not"}\n',
    ]
    for msg in json_stream(parts):
        if 'type' in msg and msg['type'] == 'core/error':
            return
    assert False


def test_json_stream_empty():
    parts = [' \n', '  ', '\n']
    assert len([p for p in json_stream(parts)]) == 0



import asyncio
from typing import Any
from cowait.utils import Version
from ..status import WAIT, WORK, FAIL
from ..messages import TASK_INIT, TASK_STATUS, TASK_FAIL, TASK_RETURN
from ..definition import TaskDefinition
from ..remote_task import RemoteTask


class TaskManager(dict):
    """ Keeps track of the state of subtasks """

    def __init__(self, task):
        self.task = task
        self.conns = {}

        # subscribe to child task status updates
        task.node.children.on(TASK_INIT, self.on_child_init)
        task.node.children.on(TASK_STATUS, self.on_child_status)
        task.node.children.on(TASK_RETURN, self.on_child_return)
        task.node.children.on(TASK_FAIL, self.on_child_fail)
        task.node.children.on('__close', self.on_child_close)

        # forward child events to parent
        async def forward(conn, **msg):
            if msg.get('type', '__')[:2] == '__':
                return
            await task.node.parent.send(msg)
        task.node.children.on('*', forward)

    def watch(self, task, timeout=30):
        # set up init timeout check
        self.set_init_timeout(task, timeout)
        self[task.id] = task

    def set_init_timeout(self, task, timeout):
        async def timeout_check(task, timeout):
            await asyncio.sleep(timeout)
            if not task.future.done() and task.id not in self.conns.values():
                await self.emit_child_error(
                    id=task.id, error=f'{task.id} timed out before initialization')

        self.task.node.io.create_task(timeout_check(task, timeout))

    async def on_child_init(self, conn, id: str, task: dict, version: str, **msg: dict):
        # version check
        try:
            version = Version.parse(version)
            if not version.is_compatible():
                raise RuntimeError(f'{id} runs an incompatible library version: {version}')
        except Exception as e:
            await self.emit_child_error(id=task.id, error=str(e))
            await conn.close()
            if conn in self.conns:
                del self.conns[conn]
            if id in self:
                del self[id]
            return

        # register connection if its not yet known
        if conn not in self.conns:
            self.conns[conn] = id

        if id in self:
            # we have seen this task before. thats strange and might indicate an error.
            # set task connection reference
            task = self[id]
            task.conn = conn
        else:
            # this task is not known - its either a virtual task or a child of a child.
            # check if it has the running task set as its parent, if so, try to adopt it.
            taskdef = TaskDefinition.deserialize(task)
            if taskdef.parent != self.task.id:
                return

            task = RemoteTask(taskdef, self.task.cluster)
            task.conn = conn
            self[id] = task

    async def on_child_status(self, conn, id: str, status: str, **msg: dict):
        if id in self:
            task = self[id]
            task.set_status(status)

    async def on_child_return(self, conn, id: str, result: Any, result_type: any, **msg: dict):
        if id in self:
            task = self[id]
            task.set_result(result, result_type)

    async def on_child_fail(self, conn, id: str, error: str, **msg: dict):
        if id in self:
            task = self[id]
            task.set_error(error)

    async def on_child_close(self, conn, **msg: dict) -> None:
        if conn not in self.conns:
            return

        # lookup task & remove from active connections
        task_id = self.conns[conn]
        task = self[task_id]
        del self.conns[conn]

        # unset task connection reference
        task.conn = None

        # emit an error event if connection was lost before done/fail
        if task.status == WAIT or task.status == WORK:
            await self.emit_child_error(
                conn=conn, id=task.id, error=f'Lost connection to {task_id}')

    async def emit_child_error(self, id, error, conn=None):
        await self.task.node.children.emit(type=TASK_STATUS, id=id, status=FAIL, conn=conn)
        await self.task.node.children.emit(type=TASK_FAIL, id=id, conn=conn, error=error)



from cowait.storage import StorageBackends, FileFormats
from .type import Type
from .mapping import TypeAlias


@TypeAlias(*FileFormats)
class FileType(Type):
    name: str = 'File'

    def validate(self, value: any, name: str) -> None:
        if not isinstance(value, dict):
            raise TypeError(f'Expected {name} to be dict file representation')

        if 'path' not in value:
            raise TypeError(f'File {name} has no path')

        if 'fs' not in value:
            raise TypeError(f'File {name} has no path')

    def serialize(self, file) -> dict:
        return {
            'path': file.path,
            'fs': file.fs.to_json(),
        }

    def deserialize(self, value: dict):
        fs = StorageBackends.from_json(value['fs'])
        return fs.open(value['path'])



from asyncio import coroutine
from unittest.mock import Mock
from .emitter import EventEmitter


async def test_emitter():
    emitter = EventEmitter()
    cb = Mock(side_effect=coroutine(Mock()))
    emitter.on('test', cb)
    await emitter.emit('test')
    cb.assert_called()


async def test_emitter_wildcard():
    emitter = EventEmitter()
    cb = Mock(side_effect=coroutine(Mock()))
    emitter.on('*', cb)
    await emitter.emit('test')
    cb.assert_called()


async def test_emitter_off():
    emitter = EventEmitter()
    cb = Mock(side_effect=coroutine(Mock()))
    emitter.on('test', cb)
    emitter.off('test', cb)
    await emitter.emit('test')
    cb.assert_not_called()



from cowait.test import capture_task_events
from cowait.tasks import TASK_RETURN, TASK_STATUS, DONE


def test_rpc():
    # we expect to find two return values
    # and two success statuses
    output = capture_task_events('cowait.test.tasks.rpc_parent')
    assert output.count(type=TASK_RETURN) == 2
    assert output.count(type=TASK_STATUS, status=DONE) == 2



import fsspec


class StorageBackends(object):
    cache = {}

    def __init__(self, backends: dict = {}):
        self.backends = {}
        if not isinstance(backends, dict):
            raise TypeError('backends must be a dict')
        for name, store in backends.items():
            self.add(name, **store)

    def add(self, name: str, protocol: str, **storage) -> None:
        fs = fsspec.filesystem(protocol, **storage)
        self.backends[name] = fs

    def get(self, attr) -> fsspec.AbstractFileSystem:
        return self.backends.get(attr)

    def __getattr__(self, attr) -> fsspec.AbstractFileSystem:
        return self.backends.get(attr)

    def __getitem__(self, attr) -> fsspec.AbstractFileSystem:
        return self.backends.get(attr)

    def open(self, path: str, **kwargs):
        if len(self.backends) > 1:
            raise RuntimeError('Cant use open() shorthand with multiple backends')
        if len(self.backends) == 0:
            raise RuntimeError('No storage backend defined')
        fs = next(iter(self.backends.values()))
        return fs.open(path, **kwargs)

    @staticmethod
    def from_json(json):
        if json in StorageBackends.cache:
            return StorageBackends.cache[json]

        fs = fsspec.AbstractFileSystem.from_json(json)
        StorageBackends.cache[json] = fs
        return fs



import socket


PORT = 1337


def get_local_ip():
    hostname = socket.gethostname()
    return socket.gethostbyname(hostname)


def get_local_connstr():
    local_ip = get_local_ip()
    return f'ws://{local_ip}/ws'



import numpy as np
from .type import Type
from .simple import Int, Float, Bool
from .mapping import TypeAlias


@TypeAlias(
    np.byte, np.ubyte, np.int8, np.uint8,
    np.short, np.ushort, np.int16, np.uint16,
    np.intc, np.uintc, np.int_, np.uint, np.int32, np.uint32,
    np.longlong, np.ulonglong, np.int64, np.uint64,
    np.intp, np.uintp,
)
class NumpyInt(Int):
    name = 'NumpyInt'

    def deserialize(self, value: int) -> np.int64:
        return np.int64(value)


@TypeAlias(
    np.half, np.float16,
    np.single, np.float32, np.float_,
    np.double, np.longdouble, np.float64,
)
class NumpyFloat(Float):
    name = 'NumpyFloat'

    def deserialize(self, value: float) -> np.float64:
        return np.float64(value)


@TypeAlias(np.bool_)
class NumpyBool(Bool):
    name = 'NumpyBool'

    def deserialize(self, value: bool) -> np.bool_:
        return np.bool_(value)


@TypeAlias(np.ndarray)
class NumpyArray(Type):
    name = 'NumpyArray'

    def validate(self, value: list, name: str) -> None:
        if not isinstance(value, list):
            raise ValueError(f'Expected {name} to be a list')

    def serialize(self, value: np.ndarray) -> list:
        return value.tolist()

    def deserialize(self, value: list) -> np.ndarray:
        return np.array(value)



import traceback
from datetime import datetime
from cowait.tasks import Task, TaskError, rpc, sleep
from cowait.utils import parse_task_image_name
from .schedule_definition import ScheduleDefinition


class ScheduleTask(Task):
    async def run(self, schedule, target, **inputs):
        image, name = parse_task_image_name(target)

        # instantiate schedule
        self.schedule = ScheduleDefinition(schedule)

        print(f'Scheduled {name} ({image}) at {schedule}')
        print(f'The current time is', datetime.now())

        last_run = datetime(1970, 1, 1)
        while True:
            await sleep(0.5)

            now = datetime.now()
            if self.schedule.is_now():
                if self.schedule.is_at(last_run) \
                        and last_run.minute == now.minute:
                    continue

                last_run = now
                await self.execute(name, image, inputs)

    async def execute(self, name, image, inputs):
        now = datetime.now()
        print(f'{now} | Execution triggered')

        try:
            id = f'{name}-' + \
                 f'{now.year}{now.month:02d}{now.day:02d}' + \
                 f'{now.hour:02d}{now.minute:02d}'
            result = await self.spawn(
                id=id,
                name=name,
                image=image,
                inputs=inputs,
                owner=self.owner,
                env=self.env,
                meta=self.meta,
            )

            end = datetime.now()
            duration = end - now
            print(f'{end} | Finished in {duration}. Result:')
            print(result)

        except TaskError:
            end = datetime.now()
            duration = end - now
            print(f'{end} | Error after {duration}. Exception:')
            traceback.print_exc()

    @rpc
    async def reschedule(self, schedule):
        # instantiate/validate schedule
        self.schedule = ScheduleDefinition(schedule)
        return schedule



# flake8: noqa: F401

from .type import Type
from .dict import Dict
from .list import List
from .custom import CustomType
from .simple import Any, String, Int, Float, Bool, DateTime, Void
from .numpy import NumpyInt, NumpyFloat, NumpyBool, NumpyArray
from .fs import FileType
from .mapping import TypeAlias
from .utils import typed_arguments, typed_return, \
    typed_call, typed_async_call, serialize, deserialize, \
    get_parameter_types, get_parameter_defaults, get_return_type, \
    type_from_description



from cowait.network import Conn
from cowait.tasks import TaskDefinition
from cowait.tasks.messages import TASK_INIT, TASK_STATUS, TASK_RETURN, TASK_FAIL, TASK_LOG


class TaskList(dict):
    """ In-memory database containing all seen tasks and logs """

    def __init__(self, task):
        task.node.children.on(TASK_INIT, self.on_init)
        task.node.children.on(TASK_STATUS, self.on_status)
        task.node.children.on(TASK_RETURN, self.on_return)
        task.node.children.on(TASK_FAIL, self.on_fail)
        task.node.children.on(TASK_LOG, self.on_log)

    async def on_init(self, conn: Conn, id: str, task: dict, **msg):
        self[id] = TaskDefinition.deserialize(task)

    async def on_status(self, conn: Conn, id, status, **msg):
        if id in self:
            self[id].status = status

    async def on_fail(self, conn: Conn, id, error, **msg):
        if id in self:
            self[id].error = error

    async def on_return(self, conn: Conn, id, result, **msg):
        if id in self:
            self[id].result = result

    async def on_log(self, conn: Conn, id, file, data, **msg):
        if id in self:
            if not hasattr(self[id], 'log'):
                self[id].log = ''
            self[id].log += data



import pytest
from datetime import datetime
from .schedule_definition import ScheduleDefinition, schedule_match


def test_is_value():
    assert schedule_match("*", 1, 0, 100)

    assert schedule_match("1", 1, 0, 100)
    assert not schedule_match("1", 2, 0, 100)

    assert schedule_match("0-2", 1, 0, 100)
    assert not schedule_match("0-2", 3, 0, 100)

    assert schedule_match("*/5", 5, 0, 100)
    assert not schedule_match("*/5", 6, 0, 100)

    with pytest.raises(ValueError):
        schedule_match("61 * * * *", 1, 0, 59)

    with pytest.raises(ValueError):
        schedule_match("1 * * * *", 1, 10, 59)


def test_is_schedule():
    def is_schedule(schedule, time):
        s = ScheduleDefinition(schedule)
        return s.is_at(time)

    assert is_schedule("* * * * *", datetime.now())

    # invalid format
    with pytest.raises(ValueError):
        is_schedule("* **", datetime.now())

    assert is_schedule("*/3 21 * * *", datetime(2000, 1, 1, 21, 3))
    assert not is_schedule("*/3 21 * * *", datetime(2000, 1, 1, 22, 3))
    assert not is_schedule("*/3 21 * * *", datetime(2000, 1, 1, 21, 4))



import os
import os.path
import pytest
from .context import CowaitContext


def test_create_context():
    context = CowaitContext.open('test')
    assert context.root_path == os.path.join(os.getcwd(), 'test')

    # raise error if path is invalid
    with pytest.raises(ValueError):
        context = CowaitContext.open('does_not_exist')

    # raise error if no context definition is found
    local = CowaitContext.open()
    assert local.root_path == os.getcwd()


def test_get_files():
    context = CowaitContext.open('test')

    # context files
    path_abs = context.file('Dockerfile')
    path_rel = context.file_rel('Dockerfile')
    assert path_abs == os.path.join(context.root_path, context.get('workdir', '.'), 'Dockerfile')
    assert path_rel == 'Dockerfile'



from __future__ import annotations
from abc import abstractmethod
from typing import Iterable
from cowait.tasks import TaskDefinition, RemoteTask
from cowait.utils import EventEmitter
from .const import ENV_TASK_CLUSTER, ENV_TASK_DEFINITION, ENV_GZIP_ENABLED, MAX_ENV_LENGTH
from .errors import ProviderError
from .utils import env_pack


class ClusterProvider(EventEmitter):
    def __init__(self, type, args={}):
        super().__init__()
        self.type = type
        self.args = args

    @abstractmethod
    def spawn(self, taskdef: TaskDefinition) -> RemoteTask:
        """ Spawn a task in the cluster """
        raise NotImplementedError()

    @abstractmethod
    def destroy(self, task_id: str) -> None:
        """ Destroy a task """
        raise NotImplementedError()

    @abstractmethod
    def destroy_all(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def destroy_children(self, parent_id: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def wait(self, task: RemoteTask) -> bool:
        """ Wait for task to exit. Returns True on clean exit. """
        raise NotImplementedError()

    @abstractmethod
    def logs(self, task: RemoteTask) -> Iterable[str]:
        """ Stream logs from task """
        raise NotImplementedError()

    @abstractmethod
    def list_all(self) -> list:
        raise NotImplementedError()

    def serialize(self) -> dict:
        """ Serialize ClusterProvider into a dict """
        return {
            'type': self.type,
            **self.args,
        }

    def create_env(self, taskdef: TaskDefinition) -> dict:
        """
        Create a container environment dict from a task definition.

        Arguments:
            taskdef (TaskDefinition): Task definition

        Returns:
            env (dict): Environment variable dict
        """

        env = {
            **taskdef.env,
            ENV_GZIP_ENABLED:    '1',
            ENV_TASK_CLUSTER:    env_pack(self.serialize()),
            ENV_TASK_DEFINITION: env_pack(taskdef.serialize()),
        }

        length = 0
        for key, value in env.items():
            length += len(key) + len(value)
        if length > MAX_ENV_LENGTH:
            raise ProviderError(f'Task input too long. Was {length}, max: {MAX_ENV_LENGTH}')

        return env

    def find_agent(self):
        return None



import json
import gzip
from base64 import b64encode, b64decode


def env_pack(value):
    value = json.dumps(value).encode('utf-8')
    value = gzip.compress(value)
    return b64encode(value)


def env_unpack(value):
    value = b64decode(value)
    value = gzip.decompress(value)
    return json.loads(value)



import asyncio


async def join(tasks):
    return await asyncio.gather(*tasks)
    done, pending = await asyncio.wait(tasks)
    return list(done)


async def gather(*tasks):
    return await asyncio.gather(*tasks)
    done, pending = await asyncio.wait(tasks)
    return list(done)



import pytest
from .settings_dict import SettingsDict


def test_get_values():
    settings = SettingsDict(data={
        'repo': 'repo',
        'cluster': {
            'type': 'docker',
        },
    })
    assert settings['repo'] == 'repo'
    assert settings['cluster.type'] == 'docker'

    # unset keys with a provided default should return the default
    assert settings.get('cluster.undefined', 'nothing') == 'nothing'

    # uset parent with a provided default should return the default
    assert settings.get('undefined.undefined', 'nothing') == 'nothing'

    assert settings.has('repo')
    assert settings.has('cluster.type')

    # unset keys without a default should raise errors
    with pytest.raises(KeyError):
        assert settings.get('unset')

    # unset child keys with no default should raise errors
    with pytest.raises(KeyError):
        assert settings['cluster.undefined']

    # unset child keys with unset parent no default should raise errors
    with pytest.raises(KeyError):
        assert settings['undefined.undefined']


def test_set_values():
    settings = SettingsDict(data={})
    settings['nested.key'] = 1
    assert isinstance(settings.data['nested'], dict)
    assert settings.data['nested']['key'] == 1

    settings['nested.key2'] = 2
    assert settings.data['nested'] == {'key': 1, 'key2': 2}



from cowait.test import capture_task_events
from cowait.tasks import TASK_INIT, TASK_STATUS, TASK_FAIL, TASK_RETURN, WORK, DONE, FAIL


def test_task_success():
    """ Ensures that successful task executions emit all the basic events """
    output = capture_task_events('cowait.test.tasks.utility', inputs={'return': 'everything ok'})
    assert len(output.unique('id')) == 1
    assert output.has(type=TASK_INIT)
    assert output.has(type=TASK_STATUS, status=WORK)
    assert output.has(type=TASK_STATUS, status=DONE)
    assert output.has(type=TASK_RETURN, result='everything ok')
    assert not output.has(type=TASK_FAIL)
    assert not output.has(type=TASK_STATUS, status=FAIL)


def test_task_error():
    """ Ensures that failed task executions emit all the basic events """
    output = capture_task_events('cowait.test.tasks.utility', inputs={'error': True})
    assert len(output.unique('id')) == 1
    assert output.has(type=TASK_INIT)
    assert output.has(type=TASK_FAIL, error='test error')
    assert output.has(type=TASK_STATUS, status=FAIL)
    assert not output.has(type=TASK_RETURN)
    assert not output.has(type=TASK_STATUS, status=DONE)



import dill
from base64 import b64encode, b64decode


def pickle_task(task):
    serialized = dill.dumps(task)
    return str(b64encode(serialized), encoding='utf-8')


def unpickle_task(task):
    funcbytes = b64decode(task)
    return dill.loads(funcbytes)



from .dict import Dict
from .simple import Any, Int
from .utils import get_return_type, get_parameter_types, get_parameter_defaults


def test_get_parameter_defaults():
    def test_func(a, b=1, c='hi'):
        pass

    defaults = get_parameter_defaults(test_func)
    assert 'a' not in defaults
    assert 'b' in defaults and defaults['b'] == 1
    assert 'c' in defaults and defaults['c'] == 'hi'


def test_get_parameter_types():
    def test_func(a, b: int, c: any):
        pass

    types = get_parameter_types(test_func)
    assert isinstance(types, Dict)

    shape = types.shape
    assert 'a' in shape and isinstance(shape['a'], Any)
    assert 'b' in shape and isinstance(shape['b'], Int)
    assert 'c' in shape and isinstance(shape['c'], Any)


def test_get_return_type():
    def returns_int() -> int:
        pass

    def returns_any():
        pass

    return_type = get_return_type(returns_int)
    assert isinstance(return_type, Int)

    return_type = get_return_type(returns_any)
    assert isinstance(return_type, Any)



import re


class EventLog(list):
    """
    EventList is used to organize and query task event output.
    """

    def collect(self, stream: iter):
        """ Collects all the output events from a task. """
        for msg in stream:
            self.append(msg)

    def match(self, **search: dict) -> list:
        """
        Returns all events with fields matching the provided regex dict.
        Each value in the search dict is matched against the event value at the corresponding key.
        """
        return EventLog([d for d in self if dict_match(search, d)])

    def match_one(self, **search: dict) -> dict:
        """
        Returns the first event fields matching the provided regex dict.
        Each value in the search dict is matched against the event value at the corresponding key.
        """
        results = self.match(**search)
        if len(results) == 0:
            return None
        return results[0]

    def count(self, **search: dict) -> int:
        matches = 0
        for item in self:
            if dict_match(search, item):
                matches += 1
        return matches

    def has(self, **search: dict) -> bool:
        """
        Checks if a event of a certain type has been captured with fields matching the provided
        regex dict. Each value in the search dict is matched against the event value at the
        corresponding key.
        """
        return self.count(**search) > 0

    def extract(self, key: str) -> list:
        return [msg[key] for msg in self]

    def unique(self, key) -> set:
        return set(self.extract(key))


def dict_match(search: dict, item: dict) -> bool:
    for key, regex in search.items():
        if regex is None:
            continue
        if key not in item:
            return False
        if re.search(regex, item[key]) is None:
            return False
    return True



import json
import numpy as np
from .utils import serialize, deserialize


def test_serialize_np_int():
    i = np.int64(3)
    data = serialize(i)
    assert data == 3

    # should be possible to serialize
    json.dumps(data)

    i2 = deserialize(data, 'NumpyInt')
    assert isinstance(i2, np.int64)
    assert i2 == 3


def test_serialize_np_float():
    f = np.float64(3.14)
    data = serialize(f)
    assert data == 3.14

    # should be possible to serialize
    json.dumps(data)

    f2 = deserialize(data, 'NumpyFloat')
    assert isinstance(f2, np.float64)
    assert f2 == 3.14


def test_serialize_np_bool():
    b = np.bool_(True)
    data = serialize(b)
    assert data

    # should be possible to serialize
    json.dumps(data)

    b2 = deserialize(data, 'NumpyBool')
    assert isinstance(b2, np.bool_)
    assert b2


def test_serialize_np_array():
    arr = np.array([[1, 2], [1, 2]])
    data = serialize(arr)
    assert data == [[1, 2], [1, 2]]

    # should be possible to serialize
    json.dumps(data)

    arr2 = deserialize(data, 'NumpyArray')
    assert isinstance(arr2, np.ndarray)
    assert arr2.tolist() == [[1, 2], [1, 2]]


def test_serialize_nested_np():
    data = serialize({
        'num': np.int64(3),
    })
    assert type(data['num']) == int

    # should be possible to serialize
    json.dumps(data)



from __future__ import annotations
from datetime import datetime, timezone
from marshmallow import Schema, fields, post_load
from ..utils import uuid


def generate_task_id(name: str) -> str:
    if '.' in name:
        dot = name.rfind('.')
        name = name[dot+1:]

    name = name.replace('.', '-')
    name = name.replace('_', '-')

    return '%s-%s' % (name.lower(), uuid())


class TaskDefinition(object):
    """
    Defines a Task :)

    Attributes:
        name (str): Task import name.
        image (str): Task image.
        id (str): Task id. If None, an id will be autogenerated.
        upstream (str): Upstream connection string. Defaults to None.
        inputs (dict): Input values
        meta (dict): Freeform metadata
        env (dict): Environment variables
        ports (dict): Port forwards
        routes (dict): HTTP Ingresses
        volumes (dict): List of volumes
        cpu (str): CPU allocation
        memory (str): Memory allocation
        owner (str): Owner name
        created_at (DateTime): Creation date
    """

    def __init__(
        self,
        name:      str,
        image:     str,
        id:        str = None,
        upstream:  str = None,
        parent:    str = None,
        inputs:    dict = {},
        meta:      dict = {},
        env:       dict = {},
        ports:     dict = {},
        routes:    dict = {},
        volumes:   dict = {},
        storage:   dict = {},
        cpu:       str = '0',
        memory:    str = '0',
        owner:     str = '',
        created_at: datetime = None,
    ):
        """
        Arguments:
            name (str): Task import name.
            image (str): Task image.
            id (str): Task id. If None, an id will be autogenerated.
            upstream (str): Upstream connection string. Defaults to None.
            inputs (dict): Input values
            meta (dict): Freeform metadata
            env (dict): Environment variables
            ports (dict): Port forwards
            routes (dict): HTTP Ingresses
            volumes (dict): List of volumes
            cpu (str): CPU allocation
            memory (str): Memory allocation
            owner (str): Owner name
            created_at (DateTime): Creation date
        """
        self.id = generate_task_id(name) if not id else id
        self.name = name
        self.image = image
        self.parent = parent
        self.upstream = upstream
        self.inputs = inputs
        self.meta = meta
        self.env = env
        self.ports = ports
        self.routes = routes
        self.cpu = cpu
        self.memory = memory
        self.owner = owner
        self.volumes = volumes
        self.storage = storage

        if created_at is None:
            self.created_at = datetime.now(timezone.utc)
        elif isinstance(created_at, datetime):
            self.created_at = created_at
        elif isinstance(created_at, str):
            self.created_at = datetime.fromisoformat(created_at)
        else:
            raise TypeError(f'Expected created_at to be None or datetime, got {created_at}')

    def serialize(self) -> dict:
        """ Serialize task definition to a dict """
        return TaskDefinitionSchema().dump(self)

    @staticmethod
    def deserialize(taskdef: dict) -> TaskDefinition:
        """ Deserialize task definition from a dict """
        return TaskDefinitionSchema().load(taskdef)


class TaskDefinitionSchema(Schema):
    """ TaskDefinition serialization schema. """

    id = fields.Str(required=True)
    name = fields.Str(required=True)
    image = fields.Str(required=True)
    upstream = fields.Str(allow_none=True)
    parent = fields.Str(allow_none=True)
    inputs = fields.Dict(missing={})
    meta = fields.Dict(missing={})
    env = fields.Dict(missing={})
    ports = fields.Dict(missing={})
    routes = fields.Dict(missing={})
    cpu = fields.Str(missing='0')
    memory = fields.Str(missing='0')
    owner = fields.Str(missing='')
    created_at = fields.DateTime('iso', default=lambda: datetime.now(timezone.utc))
    storage = fields.Dict(missing={})
    volumes = fields.Mapping(
        keys=fields.Str(),
        values=fields.Mapping(),
        missing={}
    )

    @post_load
    def make_taskdef(self, data: dict, **kwargs) -> TaskDefinition:
        return TaskDefinition(**data)



import pytest
from .mapping import is_cowait_type, convert_type, get_type
from .simple import Any, String, Int, Float, Bool
from .dict import Dict
from .list import List


class TestType(Any):
    """ A custom input type """
    pass


class NotType(object):
    """ Another object that is not a valid type """
    pass


def test_is_cowait_type():
    assert is_cowait_type(str)
    assert is_cowait_type(int)
    assert is_cowait_type(float)
    assert is_cowait_type(bool)
    assert is_cowait_type(dict)
    assert is_cowait_type(list)
    assert is_cowait_type(TestType())
    assert not is_cowait_type(NotType)


def test_get_input_type():
    assert isinstance(convert_type(str), String)
    assert isinstance(convert_type(int), Int)
    assert isinstance(convert_type(float), Float)
    assert isinstance(convert_type(bool), Bool)
    assert isinstance(convert_type(dict), Dict)
    assert isinstance(convert_type(list), List)


def test_string():
    string = String()
    string.validate('hello', 'OK')
    string.validate(3, 'OK')
    string.validate(3.14, 'OK')

    assert string.deserialize(3) == '3'
    assert string.deserialize(3.14) == '3.14'
    assert string.deserialize('hello') == 'hello'


def test_integer():
    integer = Int()
    integer.validate(3, 'Three')
    integer.validate('6', 'Six')

    with pytest.raises(ValueError):
        integer.validate('shit', 'StringFail')

    with pytest.raises(ValueError):
        integer.validate('1.23', 'FloatFail')

    assert integer.deserialize(3) == 3
    assert integer.serialize(3) == 3


def test_float():
    float = Float()

    float.validate(3.14, 'Pi')
    float.validate('3.14', 'Pi')
    float.validate(3, 'Three')
    float.validate('3', 'Three')

    with pytest.raises(ValueError):
        float.validate('shit', 'StringFail')

    assert float.deserialize('3.14') == 3.14
    assert float.serialize(3.14) == 3.14


def test_bool():
    bool = Bool()

    bool.validate('True', 'True')
    bool.validate('false', 'False')

    with pytest.raises(ValueError):
        bool.validate(1, 'int')

    with pytest.raises(ValueError):
        bool.validate('', 'None')

    assert bool.deserialize('true')
    assert not bool.deserialize('false')


def test_list():
    mixlist = List()
    mixlist.validate([1, 'str', 3.14], 'mix')

    # typed list
    intlist = List(Int())
    intlist.validate([1, 2], 'ok')

    with pytest.raises(ValueError):
        intlist.validate([1, 'woops'], 'fail')

    assert intlist.deserialize([1, 2]) == [1, 2]


def test_dict():
    dct = Dict({
        'text': String(),
        'number': Int(),
    })

    dct.validate({
        'text': 'hello',
        'number': 123,
    }, 'ok')

    # member missing
    with pytest.raises(ValueError):
        dct.validate({
            'text': 'hello',
        }, 'missing')

    # type error
    with pytest.raises(ValueError):
        dct.validate({
            'text': 'hello',
            'number': 'one'
        }, 'type err')


def test_composite():
    custom = Dict({
        'int_list': List(Int()),
        'subdict': Dict({
            'number': Int(),
        }),
    })

    custom.validate({
        'int_list': [1, 2],
        'subdict': {
            'number': 1,
        },
    }, 'ok')

    with pytest.raises(ValueError):
        custom.validate({
            'int_list': [1, 2],
            'subdict': {},
        }, 'subdict fail')


def test_instantiate_type():
    """ Ensures all default types can be auto instantiated """
    assert isinstance(convert_type(Any), Any)
    assert isinstance(convert_type(String), String)
    assert isinstance(convert_type(Int), Int)
    assert isinstance(convert_type(Float), Float)
    assert isinstance(convert_type(Bool), Bool)
    assert isinstance(convert_type(Dict), Dict)
    assert isinstance(convert_type(List), List)


def test_instantiate_with_non_default_params():
    class TypeWithParam(Any):
        def __init__(self, non_default):
            pass

    with pytest.raises(TypeError):
        convert_type(TypeWithParam)


def test_instantiate_with_default_params():
    class TypeWithParam(Any):
        def __init__(self, a=1):
            self.a = a

    obj = convert_type(TypeWithParam)
    assert isinstance(obj, TypeWithParam)
    assert obj.a == 1


def test_get_type():
    assert get_type('str') is String
    assert get_type('int') is Int



# flake8: noqa: F401

from .backends import StorageBackends

from fsspec.implementations.local import LocalFileOpener
from fsspec.implementations.http import HTTPFile
from fsspec.implementations.ftp import FTPFile

FileFormats = [LocalFileOpener, HTTPFile, FTPFile]

# HDFS (requires pyarrow)
try:
    from fsspec.implementations.hdfs import HDFSFile
    FileFormats.append(HDFSFile)
except ImportError:
    pass

# S3 (requires s3fs)
try:
    from s3fs import S3File
    FileFormats.append(S3File)
except ImportError:
    pass



import traceback
from cowait.engine import ClusterProvider
from cowait.tasks import Task, TaskDefinition, TaskError
from cowait.types import typed_arguments, typed_return, get_parameter_defaults
from cowait.storage import StorageBackends
from .worker_node import WorkerNode
from .loader import load_task_class


async def execute(cluster: ClusterProvider, taskdef: TaskDefinition) -> None:
    """
    Executes a task on this worker node.
    """

    # create network node
    node = WorkerNode(taskdef.id, taskdef.upstream)

    # start upstream client
    await node.connect()

    try:
        # init should always be the first command sent
        await node.parent.send_init(taskdef)

        # run task within a log capture context
        with node.capture_logs():

            # instantiate
            TaskClass = load_task_class(taskdef.name)
            task = TaskClass(
                taskdef=taskdef,
                cluster=cluster,
                node=node,
            )

            # initialize storage
            setattr(task, 'storage', StorageBackends(taskdef.storage))

            # initialize task
            task.init()

            # start http server
            # this must happen after task.init() so that tasks have a chance
            # to register extra http routes.
            node.serve()

            # unpack wrapped function if defined.
            # allows typechecking of functional tasks
            taskfunc = task.run
            if hasattr(TaskClass, '__wraps__'):
                taskfunc = TaskClass.__wraps__

            # prepare arguments
            inputs = {
                **get_parameter_defaults(taskfunc),
                **taskdef.inputs,
            }

            # before hook
            inputs = await task.before(inputs)
            if inputs is None:
                raise ValueError(
                    'Task.before() returned None, '
                    'did you forget to return the inputs?')

            # typecheck arguments
            inputs = typed_arguments(taskfunc, inputs)

            # set state to running
            await node.parent.send_run()

            # execute task
            result = await task.run(**inputs)

            # after hook
            await task.after(inputs)

            # wait for dangling tasks
            await handle_orphans(task)

            # prepare & typecheck result
            result, result_type = typed_return(taskfunc, result)

            # submit result
            await node.parent.send_done(result, result_type.describe())

    except TaskError as e:
        # pass subtask errors upstream
        await node.parent.send_fail(
            f'Caught exception in subtask {taskdef.id}:\n'
            f'{e.error}')
        raise e

    except Exception as e:
        # capture local errors
        error = traceback.format_exc()
        await node.parent.send_fail(
            f'Caught exception in {taskdef.id}:\n'
            f'{error}')
        raise e

    finally:
        await node.close()

        # clear active task
        Task.set_current(None)


async def handle_orphans(task: Task, mode: str = 'stop') -> None:
    orphans = filter(lambda child: not child.done, task.subtasks.values())
    for orphan in orphans:
        if mode == 'wait':
            print('~~ waiting for orphaned task', orphan.id)
            await orphan
        elif mode == 'stop':
            print('~~ stopping orphaned task', orphan.id)
            await orphan.stop()
        else:
            raise RuntimeError(f'Unknown orphan mode {mode}')



from .dict import Dict
from .list import List
from .custom import CustomType
from .simple import Int, String
from .mapping import TypeAlias
from .utils import type_from_description


def test_deserialize_dict_type():
    desc = type_from_description({
        'text': 'str',
        'number': 'int',
    })
    assert isinstance(desc, Dict)
    assert isinstance(desc['text'], String)
    assert isinstance(desc['number'], Int)


def test_deserialize_list_type():
    desc = type_from_description(['str'])
    assert isinstance(desc, List)
    assert isinstance(desc.item, String)


class ComplexCustom(object):
    def __init__(self, text, number):
        self.text = text
        self.number = number


@TypeAlias(ComplexCustom)
class ComplexCustomType(CustomType):
    name = 'ComplexCustomType'

    def __init__(self):
        super().__init__(ComplexCustom, {
            'text': String(),
            'number': Int(),
        })


def test_custom_type_serialization():
    obj = ComplexCustom(text='hej', number=3)
    type = ComplexCustomType()

    desc = type.describe()
    assert desc == 'ComplexCustomType'

    data = type.serialize(obj)
    assert 'text' in data and data['text'] == obj.text
    assert 'number' in data and data['number'] == obj.number

    # the custom type should be available from the type registry
    # ensure deserialization yields a proper object with the expected values
    type2 = type_from_description(type.name)
    obj2 = type2.deserialize(data)
    assert isinstance(obj2, ComplexCustom)
    assert obj2.text == obj.text
    assert obj2.number == obj.number



from cowait.tasks.pickled import PickledTask, pickle_task


def task(task_func):
    def spawn_task(*args, **inputs):
        if len(args) > 0:
            raise TypeError('Use keyword arguments for task inputs')
        return PickledTask(func=pickle_task(task_func), **inputs)
    return spawn_task



from dask.distributed import Client
from cowait.types import Type, TypeAlias

SCHEDULER_URI = 'scheduler_uri'


@TypeAlias(Client)
class DaskClientType(Type):
    name: str = 'DaskClient'

    def validate(self, client: Client, name: str) -> None:
        if not isinstance(client, dict):
            raise ValueError(f'{name} is not a DaskClient')
        if SCHEDULER_URI not in client:
            raise ValueError(f'{name} is not a valid DaskClient: Missing {SCHEDULER_URI}')

    def serialize(self, client: Client):
        return {
            SCHEDULER_URI: client.scheduler.address,
        }

    def deserialize(self, value: dict):
        return Client(
            address=value[SCHEDULER_URI]
        )




class EventEmitter(object):
    def __init__(self):
        self.callbacks = {'*': []}

    def on(self, type: str, callback: callable) -> None:
        try:
            self.callbacks[type].insert(0, callback)
        except KeyError:
            self.callbacks[type] = [callback]

    def off(self, type: str, callback: callable) -> None:
        try:
            self.callbacks[type].remove(callback)
        except KeyError:
            pass

    async def emit(self, type: str, **kwargs: dict) -> None:
        if type == '*':
            raise ValueError('Cannot emit wildcard events')

        try:
            # wildcard events
            for callback in self.callbacks['*']:
                await callback(type=type, **kwargs)

            # actual event
            for callback in self.callbacks[type]:
                await callback(**kwargs)

        except KeyError:
            pass

    def emit_sync(self, type: str, **kwargs: dict) -> None:
        if type == '*':
            raise ValueError('Cannot emit wildcard events')

        try:
            # wildcard events
            for callback in self.callbacks['*']:
                callback(type=type, **kwargs)

            # actual event
            for callback in self.callbacks[type]:
                callback(**kwargs)

        except KeyError:
            pass



import os
import os.path
import docker
from .context import CowaitContext
from .docker_file import Dockerfile

client = docker.from_env()


class BuildError(RuntimeError):
    pass


class TaskImage(object):
    def __init__(self, context):
        self.context = context

    @property
    def name(self):
        return self.context.image

    def build(self, base: str, requirements: str = None):
        """ Build task image """

        # create temporary dockerfile
        df = Dockerfile(base)

        # install task-specific requirements
        requirements = self.context.file_rel('requirements.txt')
        if requirements:
            df.copy(f'./{requirements}', './requirements.txt')
            df.run('pip install -r ./requirements.txt')

        # copy source code
        df.copy('.', '.')

        workdir = self.context.workdir
        if workdir != '.':
            df.workdir(os.path.join('/var/task', workdir))

        return TaskImage.build_image(
            dockerfile=str(df),
            path=self.context.root_path,
            tag=f'{self.name}:latest',
        )

    def push(self):
        """
        Push context image to a remote registry.
        """
        if not self.image:
            raise RuntimeError('Task must be built first')

        self.image.tag(
            repository=self.name,
            tag='latest',
        )

        logs = client.images.push(
            repository=self.name,
            tag='latest',
            stream=True
        )
        return logs

    @staticmethod
    def open(context: CowaitContext = None):
        # automatically create context
        if context is None:
            context = CowaitContext.open()

        return TaskImage(
            context=context,
        )

    @staticmethod
    def get(name_or_id):
        return client.images.get(name_or_id)

    @staticmethod
    def build_image(**kwargs):
        logs = client.api.build(decode=True, rm=True, **kwargs)

        image_hash = None
        for log in logs:
            if 'stream' in log:
                print(log['stream'], end='', flush=True)
            elif 'aux' in log and 'ID' in log['aux']:
                image_hash = log['aux']['ID']
            elif 'errorDetail' in log:
                raise BuildError(log['errorDetail']['message'])
            else:
                print(log)

        if image_hash is not None:
            return TaskImage.get(image_hash)



import docker
import requests.exceptions
from docker.types import Mount
from cowait.tasks import TaskDefinition, RemoteTask
from cowait.utils import json_stream
from .cluster import ClusterProvider
from .const import LABEL_TASK_ID, LABEL_PARENT_ID
from .errors import ProviderError
from .routers import create_router

DEFAULT_NETWORK = 'cowait'


class DockerTask(RemoteTask):
    def __init__(
        self,
        cluster: ClusterProvider,
        taskdef: TaskDefinition,
        container
    ):
        super().__init__(
            cluster=cluster,
            taskdef=taskdef,
        )
        self.container = container
        self.ip = taskdef.id  # id should be routable within docker

    def __str__(self):
        return f'DockerTask({self.id}, {self.status}, {self.inputs}, {self.container.id[:10]})'


class DockerProvider(ClusterProvider):
    def __init__(self, args={}):
        super().__init__('docker', args)
        self.docker = docker.from_env()
        self.router = create_router(self, self.args.get('router', 'local'))

    @property
    def network(self):
        return self.args.get('network', DEFAULT_NETWORK)

    def spawn(self, taskdef: TaskDefinition) -> DockerTask:
        try:
            self.ensure_network()

            self.emit_sync('prepare', taskdef=taskdef)

            container = self.docker.containers.run(
                detach=True,
                image=taskdef.image,
                name=taskdef.id,
                hostname=taskdef.id,
                network=self.network,
                ports=self.create_ports(taskdef),
                environment=self.create_env(taskdef),
                mounts=self.create_mounts(taskdef.volumes),
                labels={
                    LABEL_TASK_ID: taskdef.id,
                    LABEL_PARENT_ID: taskdef.parent,
                    **taskdef.meta,
                },
            )

            # print('~~ created docker container with id',
            #   container.id[:12], 'for task', taskdef.id)

            task = DockerTask(self, taskdef, container)
            self.emit_sync('spawn', task=task)
            return task

        except docker.errors.APIError as e:
            raise ProviderError(e.explanation)

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def list_all(self) -> list:
        """ Returns a list of all running tasks """
        try:
            containers = self.docker.containers.list(
                filters={
                    'label': LABEL_TASK_ID,
                },
            )
            return list(map(lambda c: c.labels[LABEL_TASK_ID], containers))

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def destroy_all(self) -> None:
        """ Destroys all running tasks """
        try:
            containers = self.docker.containers.list(
                all=True,
                filters={
                    'label': LABEL_TASK_ID,
                },
            )

            for container in containers:
                container.remove(force=True)

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def find_child_containers(self, parent_id: str) -> list:
        """ Finds all child containers of a given task id """
        try:
            return self.docker.containers.list(
                filters={
                    'label': f'{LABEL_PARENT_ID}={parent_id}',
                },
            )

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def destroy_children(self, parent_id: str) -> list:
        """ Destroy all child tasks of a given task id """
        try:
            children = self.find_child_containers(parent_id)

            tasks = []
            for child in children:
                tasks += self.destroy(child.labels[LABEL_TASK_ID])

            return tasks

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def destroy(self, task_id):
        """ Destroy a specific task id and all its descendants """

        # optimization: grab a list of all tasks at once, instead of querying
        # for every child.

        def kill_family(container):
            container_task_id = container.labels[LABEL_TASK_ID]
            # print(f'~~ kill {container_task_id} ({container.short_id})')

            children = self.find_child_containers(container_task_id)
            kills = []
            for child in children:
                kills += kill_family(child)

            try:
                container.remove(force=True)
            except docker.errors.APIError as e:
                if 'already in progress' in str(e):
                    pass
                else:
                    raise e
            except docker.errors.NotFound:
                pass

            kills.append(task_id)
            return kills

        try:
            container = self.docker.containers.get(task_id)
            return kill_family(container)

        except docker.errors.NotFound:
            return [task_id]

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def logs(self, task: DockerTask):
        """ Stream task logs """
        if not isinstance(task, DockerTask):
            raise TypeError('Expected a valid DockerTask')

        try:
            return json_stream(task.container.logs(stream=True))

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def wait(self, task: DockerTask) -> bool:
        """ Wait for a task to finish. Returns True on clean exit """
        result = task.container.wait()
        return result['StatusCode'] == 0

    def ensure_network(self):
        try:
            self.docker.networks.get(self.network)
        except docker.errors.NotFound:
            print('~~ creating docker network', self.network)
            self.docker.networks.create(
                name=self.network,
                check_duplicate=False,
                driver='bridge',
                labels={
                    'cowait': '1',
                })

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def create_ports(self, taskdef):
        return taskdef.ports

    def find_agent(self):
        try:
            container = self.docker.containers.get('agent')
            if container.status != 'running':
                return None

            token = container.labels['http_token']
            return f'ws://agent/ws?token={token}'

        except docker.errors.NotFound:
            return None

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def create_mounts(self, volumes: dict) -> list:
        mounts = [
            # this is the secret sauce that allows us to create new
            # tasks as sibling containers
            create_bind_mount('/var/run/docker.sock', {
                'src': '/var/run/docker.sock',
                'mode': 'ro',
            })
        ]

        for target, volume in volumes.items():
            if 'bind' in volume:
                mounts.append(create_bind_mount(target, volume['bind']))
            elif 'tmpfs' in volume:
                mounts.append(create_tmpfs_mount(target, volume['tmpfs']))
            else:
                # print(f'!! unsupported volume: {target}')
                pass

        return mounts


def create_bind_mount(target: str, bind) -> Mount:
    mode = 'rw'
    src = None
    if isinstance(bind, str):
        src = bind
    elif isinstance(bind, dict):
        src = bind.get('src')
        mode = bind.get('mode', 'rw')
    else:
        raise TypeError(f'Invalid bind volume definition {target}')

    return Mount(
        type='bind',
        target=target,
        source=src,
        read_only=mode != 'rw',
    )


def create_tmpfs_mount(target: str, tmpfs) -> Mount:
    size = 0
    mode = 0
    if isinstance(tmpfs, dict):
        size = tmpfs.get('size')
        mode = tmpfs.get('mode', 1777)
    else:
        raise TypeError(f'Invalid tmpfs volume definition {target}')

    return Mount(
        type='tmpfs',
        target=target,
        tmpfs_size=size,
        tmpfs_mode=mode,
    )



from .run import run as run_cmd


def notebook(config, build):
    return run_cmd(
        config=config,
        task='cowait.notebook',
        build=build,
        routes={
            '/': '8888',
        },
    )



from .event_log import EventLog


def test_push_output():
    output = EventLog()
    output.append({'type': 'test'})
    output.append({'type': 'test'})
    assert len(output) == 2
    assert output.has(type='test')

    output.collect([{'type': 'one'}, {'type': 'two'}])
    assert output.has(type='one')
    assert output.has(type='two')


def test_output_matching():
    output = EventLog()
    output.append({'type': 'test', 'field': 'hello team'})
    output.append({'type': 'other', 'field': 'hello team'})
    output.append({'type': 'test', 'field': 'wut'})
    output.append({'type': 'test'})
    assert output.count(type='test', field='tea.') == 1
    assert output.count(field='tea.') == 2
    assert len(output.match(type='test', field='tea.')) == 1



from cowait.tasks.shell import ShellTask


class SparkMaster(ShellTask):
    async def before(self, inputs: dict) -> dict:
        inputs['command'] = ' '.join([
            'spark-class',
            'org.apache.spark.deploy.master.Master',
        ])
        return inputs



import sys
import json
from ..task_image import TaskImage
from ..const import CONTEXT_FILE_NAME
from .build import build


def push() -> TaskImage:
    image = build()

    if image is None or '/' not in image.name:
        print(f'Error: You must specify a full image name in '
              f'{CONTEXT_FILE_NAME} before you can push')
        return

    sys.stdout.write('pushing...')
    logs = image.push()
    progress = {}
    for log in logs:
        rows = log.decode('utf-8').split('\r\n')
        for data in rows:
            if len(data) == 0:
                continue
            update = json.loads(data.strip())
            if 'id' in update and 'progressDetail' in update:
                id = update['id']
                progress[id] = update['progressDetail']

        current = 0
        total = 0
        for detail in progress.values():
            current += detail.get('current', 0)
            total += detail.get('total', 0)

        if total > 0:
            pct = 100 * min(current / total, 1.0)
            sys.stdout.write(f'\rpushing... {pct:0.2f}%  ')
            sys.stdout.flush()

    sys.stdout.write(f'\rpushing... done       \n')
    sys.stdout.flush()
    return image



import os.path
import docker
import docker.errors
import docker.credentials.errors
from cowait.utils.const import DEFAULT_BASE_IMAGE
from ..task_image import TaskImage, BuildError
from ..docker_file import Dockerfile
from ..context import CowaitContext, CONTEXT_FILE_NAME
from ..logger import Logger


def build(quiet: bool = False, workdir: str = None) -> TaskImage:
    logger = Logger(quiet)
    try:
        if not CowaitContext.exists():
            logger.println(f'No {CONTEXT_FILE_NAME} found. '
                           f'Using default image: {DEFAULT_BASE_IMAGE}')
            return TaskImage.get(DEFAULT_BASE_IMAGE)

        context = CowaitContext.open()
        context.override('workdir', workdir)

        image = TaskImage.open(context)
        logger.header('BUILD')
        logger.println('Image:', image.name)
        logger.println('Context Root:', context.root_path)
        logger.println('Workdir:', context.workdir)

        # find task-specific requirements.txt
        # if it exists, it will be copied to the container, and installed
        requirements = context.file_rel('requirements.txt')
        if requirements:
            logger.println('* Found custom requirements.txt')

        # find custom Dockerfile
        # if it exists, build and extend that instead of the default base image
        base_image = context.base
        dockerfile = context.file('Dockerfile')
        if dockerfile:
            logger.println('* Found custom Dockerfile:', context.relpath(dockerfile))
            logger.header('BASE')

            basedf = Dockerfile.read(dockerfile)
            base = TaskImage.build_image(
                path=os.path.dirname(dockerfile),
                dockerfile=str(basedf),
            )
            if base is None:
                raise RuntimeError('Failed to build base image')
            base_image = base.id

        logger.header('IMAGE')
        image = image.build(
            base=base_image,
            requirements=requirements,
        )

        return image

    except docker.errors.DockerException as e:
        logger.print_exception(f'Docker exception: {e}')

    except BuildError as e:
        logger.print_exception(str(e))



import os.path
from cowait.engine import get_cluster_provider
from .const import CONTEXT_FILE_NAME
from .settings_dict import SettingsDict


def get_config_path():
    home = os.path.expanduser('~')
    return f'{home}/.{CONTEXT_FILE_NAME}'


class CowaitConfig(SettingsDict):
    def __init__(
        self, *,
        path: str = None,
        data: dict = None
    ):
        super().__init__(path=path, data=data)
        self.set('clusters', {
            'docker': {'type': 'docker'},
            'kubernetes': {'type': 'kubernetes'},
            **self.get('clusters', {}),
        })

    @property
    def clusters(self) -> list:
        return self.get('clusters', {}, False)

    @property
    def default_cluster(self) -> str:
        return self.get('default_cluster', 'docker', False)

    def get_cluster(self, cluster_name: str = None):
        if cluster_name is None:
            cluster_name = self.default_cluster
        if cluster_name not in self.clusters:
            raise RuntimeError(
                f'No configuration found for cluster {cluster_name}')
        return get_cluster_provider(**self.get(f'clusters.{cluster_name}'))

    @staticmethod
    def load(path: str = None) -> None:
        if path is None:
            path = get_config_path()

        if not os.path.exists(path):
            return CowaitConfig(data={})

        return CowaitConfig(path=path)

    def write(self, path: str = None) -> None:
        if path is None:
            path = get_config_path()
        return super().write(path)



from cowait.tasks import Task, TaskDefinition
from cowait.engine import DockerProvider
from cowait.utils.const import DEFAULT_BASE_IMAGE
from .event_log import EventLog


def get_test_provider():
    """ Returns an instance of the test cluster provider """
    # todo: pick provider based on env or something?
    # must be able to run all tests on different providers
    return DockerProvider()


def spawn_test_task(name, **taskdef: dict) -> Task:
    """ Spawns a task using the test cluster provider """
    provider = get_test_provider()
    return provider.spawn(TaskDefinition(**{
        'name': name,
        'image': DEFAULT_BASE_IMAGE,
        **taskdef
    }))


def capture_task_events(name, **taskdef: dict) -> EventLog:
    output = EventLog()
    task = spawn_test_task(name, **taskdef)
    output.collect(task.logs())
    return output



import inspect
import importlib
from typing import TypeVar
from cowait.tasks import Task, TaskNotFoundError


def load_task_class(task_name: str) -> TypeVar:
    """
    Loads a task class from the given module.

    Arguments:
        task_name (str): Import path of the module containing the task.

    Raises:
        TaskNotFoundError

    Returns:
        Class (type)
    """

    def is_task(obj):
        """
        Predicate function which returns true for objects that are
        subclasses of Task and defined in the loaded module.
        """
        return inspect.isclass(obj) and \
            issubclass(obj, Task) and \
            obj.__module__.startswith(task_name)

    if task_name.startswith('.') or task_name.endswith('.'):
        raise TaskNotFoundError(f'Illegal task name {task_name}')

    try:
        module = importlib.import_module(task_name)
        classes = inspect.getmembers(module, is_task)

        # return an error if no tasks were found
        if len(classes) == 0:
            raise TaskNotFoundError(
                f'Module {task_name} does not contain any tasks')

        # each module should only define one task
        if len(classes) > 1:
            raise TaskNotFoundError(
                f'Module {task_name} contains multiple task definitions')

        # return task class
        _, Class = classes[0]
        return Class

    except ModuleNotFoundError:
        raise TaskNotFoundError(f'Task module {task_name} not found')



from .definition import TaskDefinition


def test_taskdef_serialization():
    sample = {
        'id':         'lazy-kcjvnrsg',
        'image':      'cowait/task:lazy',
        'name':       'lazy',
        'parent':     'root',
        'env':        {},
        'meta':       {},
        'inputs':     {},
        'ports':      {},
        'routes':     {},
        'volumes':    {},
        'storage':    {},
        'upstream':   None,
        'created_at': '2020-02-02T20:00:02+00:00',
        'cpu':        '0',
        'memory':     '0',
        'owner':      'santa',
    }

    taskdef = TaskDefinition.deserialize(sample)
    output = taskdef.serialize()
    assert sample == output


def test_taskdef_default_id():
    """
    Task definitions should be assigned an autogenerated ID if not provided.
    """
    taskdef = TaskDefinition(
        name='test',
        image='image',
    )
    assert isinstance(taskdef.id, str)


def test_taskdef_input_id():
    """ Task definitions can be passed an ID """
    taskdef = TaskDefinition(
        id='123',
        name='test',
        image='image',
    )
    assert taskdef.id == '123'



from io import StringIO
# from unittest.mock import Mock
from .stream_capture import StreamCapture


def test_write():
    output = StringIO()
    cap = StreamCapture(output)

    cap.write('hello')
    assert cap.getvalue() == 'hello'
    assert cap.getvalue() == output.getvalue()


# def test_newline_flush():
#     cb = Mock()
#     output = Mock()
#     cap = StreamCapture(output, callback=cb)

#     cap.write('hello ')
#     assert cb.call_count == 0

#     cap.write('team\n')
#     assert cb.call_count == 1
#     cb.assert_called_with('hello team\n')

#     cap.write('again\nyet again\nye')
#     assert cb.call_count == 2
#     cb.assert_called_with('again\nyet again\nye')

#     assert output.flush.call_count == 2



import json
import time
import pytest
import docker
from .docker import DockerProvider, LABEL_TASK_ID, LABEL_PARENT_ID
from cowait.tasks import TaskDefinition

TEST_IMAGE = 'cowait/task'
TEST_TASK = 'cowait.test.tasks.utility'


def test_create_docker_task():
    dp = DockerProvider()
    docker = dp.docker

    env_vars = {'hello': 'team'}

    taskdef = TaskDefinition(
        name=TEST_TASK,
        image=TEST_IMAGE,
        parent='parent',
        env=env_vars,
        inputs={
            'hello': '123',
            'child': False,
        },
    )

    # run task
    task = dp.spawn(taskdef)

    assert task.id == taskdef.id
    assert hasattr(task, 'container')
    assert hasattr(task.container, 'id')

    # try to grab the container from docker api
    container = docker.containers.get(task.container.id)
    assert task.container == container

    # make sure container is properly labeled
    assert container.labels[LABEL_TASK_ID] == task.id
    assert container.labels[LABEL_PARENT_ID] == 'parent'

    # test task will dump info as json, so we can pick it up
    # make sure it matches what we put in.
    task_dump = None
    for msg in task.logs():
        print(msg)
        if msg['type'] == 'task/log':
            task_dump = json.loads(msg['data'])

    # wait for container to execute
    result = container.wait()
    assert result['StatusCode'] == 0

    # taskdef
    assert task_dump is not None
    assert taskdef.serialize() == task_dump['taskdef']

    # actual environment variables
    for key, val in env_vars.items():
        assert task_dump['env'][key] == val


def test_kill_docker_task():
    dp = DockerProvider()

    task = dp.spawn(TaskDefinition(
        name=TEST_TASK,
        image=TEST_IMAGE,
        inputs={'forever': True},
    ))

    # ensure container exists
    dp.docker.containers.get(task.container.id)

    # destroy it
    dp.destroy(task.id)

    # ensure it no longer exists
    with pytest.raises(docker.errors.NotFound):
        dp.docker.containers.get(task.container.id)


def test_docker_child_task():
    dp = DockerProvider()

    task = dp.spawn(TaskDefinition(
        name=TEST_TASK,
        image=TEST_IMAGE,
        inputs={'child': True},
    ))

    # wait for the child to spawn
    child = None
    for i in range(0, 10):
        children = dp.find_child_containers(task.id)
        if len(children) > 0:
            child = children[0]
            break
        time.sleep(0.5)

    # make sure we got a child
    assert child is not None

    # test list tasks
    tasks = dp.list_all()
    assert task.id in tasks
    assert child.labels[LABEL_TASK_ID] in tasks

    # kill the whole family
    dp.destroy(task.id)

    children = dp.find_child_containers(task.id)
    assert len(children) == 0


def test_docker_task_error():
    dp = DockerProvider()

    task = dp.spawn(TaskDefinition(
        name=TEST_TASK,
        image=TEST_IMAGE,
        inputs={'error': True},
    ))

    container = dp.docker.containers.get(task.container.id)
    assert task.container == container

    result = container.wait()
    assert result['StatusCode'] != 0


def test_docker_child_error():
    dp = DockerProvider()

    task = dp.spawn(TaskDefinition(
        name=TEST_TASK,
        image=TEST_IMAGE,
        inputs={'child_error': True},
    ))

    container = dp.docker.containers.get(task.container.id)
    assert task.container == container

    # child error should cause the parent to fail
    result = container.wait()
    assert result['StatusCode'] != 0



import os.path
from aiohttp import web

DASH_ROOT = '/var/cowait/cloud/build'


class Dashboard(object):
    def routes(self):
        if not self.is_built():
            print('~~ notice: dashboard not included in build')
            return []

        return [
            web.static('/static/', f'{DASH_ROOT}/static'),
            web.get('/{tail:.*}', self.get_dashboard),
        ]

    def get_dashboard(self, req):
        return web.FileResponse(f'{DASH_ROOT}/index.html')

    @staticmethod
    def is_built():
        return os.path.exists(DASH_ROOT)



import os
import asyncio
import aiohttp
from cowait.utils import EventEmitter
from .rpc_client import RpcClient


class Client(EventEmitter):
    def __init__(self):
        super().__init__()
        self.ws = None
        self.rpc = None
        self.buffer = []

    @property
    def connected(self) -> bool:
        return self.ws is not None

    async def connect(
        self,
        url: str,
        token: str,
        max_retries: int = 5
    ) -> None:
        retries = 0
        while retries < max_retries or max_retries == 0:
            retries += 1
            try:
                await self._connect(url, token)
            except aiohttp.ClientError as e:
                print('Upstream connection failed:', str(e))
                await asyncio.sleep(retries * 3)
            finally:
                self.ws = None

        if retries >= max_retries and max_retries > 0:
            # Reached maximum number of connection attempts
            # Crash the client
            print('Max upstream connection attempts reached')
            os._exit(1)

    async def _connect(self, url: str, token: str) -> None:
        headers = {'Authorization': f'Bearer {token}'}
        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(url, headers=headers) as ws:
                self.ws = ws
                self.rpc = RpcClient(ws)

                # send buffered messages
                for msg in self.buffer:
                    await self.send(msg)
                self.buffer = []

                # client loop
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        event = msg.json()

                        if self.rpc.intercept_event(**event):
                            continue

                        await self.emit(**event, conn=self)

                    elif msg.type == aiohttp.WSMsgType.CLOSE:
                        print('parent ws clean close')

                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        print('parent ws client error:', ws.exception())

    async def close(self):
        if self.connected:
            await self.ws.close()
            self.ws = None

    async def send(self, msg: dict) -> None:
        if not self.connected:
            # buffer messages while disconnected
            self.buffer.append(msg)
            return

        await self.ws.send_json(msg)



from ..config import CowaitConfig
from cowait.engine import ProviderError


def destroy(config: CowaitConfig) -> None:
    try:
        cluster = config.get_cluster()
        cluster.destroy_all()

    except ProviderError as e:
        print('Provider error:', str(e))


def list_tasks(config: CowaitConfig) -> None:
    try:
        cluster = config.get_cluster()
        tasks = cluster.list_all()
        for task in tasks:
            print(task)

    except ProviderError as e:
        print('Provider error:', str(e))


def kill(config: CowaitConfig, task_id: str):
    try:
        cluster = config.get_cluster()
        cluster.destroy(task_id)

    except ProviderError as e:
        print('Provider error:', str(e))



# flake8: noqa: F401

from .uuid import uuid
from .stream_capture import StreamCapturing
from .emitter import EventEmitter
from .names import parse_task_image_name
from .json_stream import json_stream
from .version import Version, version_string



import asyncio
from typing import Any
from cowait.network import Client
from cowait.tasks import TaskDefinition, WORK, DONE, STOP, FAIL
from cowait.tasks.messages import TASK_INIT, TASK_LOG, TASK_STATUS, TASK_RETURN, TASK_FAIL
from cowait.utils import version_string
from .logger import Logger, JSONLogger


class ParentClient(Client):
    """
    Upstream API client.
    """

    def __init__(self, id, io_loop, logger: Logger = None):
        super().__init__()
        self.id = id
        self.io = io_loop
        self.logger = logger or JSONLogger()

    async def connect(self, url: str, token: str = None) -> None:
        if url is None:
            return

        if token is None:
            token = self.id

        # super.connect() blocks for the duration of the connection
        # run it on the I/O loop
        self.io.create_task(super().connect(url, token))

        # block caller until connected
        while not self.connected:
            await asyncio.sleep(0.1)

    async def msg(self, type: str, **msg) -> None:
        """
        Send a message upstream.

        Arguments:
            type (str): Message type
            kwargs (dict): Message fields
        """
        await self.send({
            'id': self.id,
            'type': type,
            **msg,
        })

    async def send(self, msg: dict):
        # send messages on the I/O loop
        self.io.create_task(self.logger.handle(msg))
        if self.connected:
            await self.io.create_task(super().send(msg))

    async def send_init(self, taskdef: TaskDefinition) -> None:
        """
        Send a task initialization message.

        Arguments:
            taskdef (TaskDefinition): New task definition
        """
        await self.msg(
            TASK_INIT,
            task=taskdef.serialize(),
            version=version_string(),
        )

    async def send_run(self) -> None:
        """ Send status update: Running """
        await self.msg(TASK_STATUS, status=WORK)

    async def send_stop(self, id: str = None) -> None:
        """ Send status update: Stopped """
        id = self.id if id is None else id
        await self.msg(TASK_STATUS, status=STOP, id=id)

    async def send_done(self, result: Any, result_type: str = 'any') -> None:
        """
        Send status update: Done, and return a result.

        Arguments:
            result (any): Any serializable data to return to the upstream task.
            result_type (str): Result type description
        """
        await self.msg(TASK_RETURN, result=result, result_type=result_type)
        await self.msg(TASK_STATUS, status=DONE)

    async def send_fail(self, error: str) -> None:
        """
        Send an error.

        Arguments:
            error (str): Error message
        """
        await self.msg(TASK_FAIL,   error=error)
        await self.msg(TASK_STATUS, status=FAIL)

    async def send_log(self, file: str, data: str) -> None:
        """
        Send captured log output.

        Arguments:
            file (str): Capture source (stdout/stderr)
            data (str): Captured output data
        """
        await self.msg(TASK_LOG, file=file, data=data)



import asyncio
from cowait.network import Server, HttpServer, get_local_ip
from cowait.utils import StreamCapturing
from .io_thread import IOThread
from .logger import Logger
from .parent_client import ParentClient


class WorkerNode(object):
    def __init__(self, id: str, upstream: str, port: int = 80, logger: Logger = None):
        super().__init__()
        self.id = id
        self.io = IOThread()
        self.io.start()
        self.upstream = upstream
        self.http = HttpServer(port=port)
        self.children = Server(self)
        self.parent = ParentClient(id, self.io, logger)

    async def connect(self) -> None:
        await self.parent.connect(self.upstream)

    def serve(self):
        self.io.create_task(self.http.serve())

    def get_url(self):
        local_ip = get_local_ip()
        return f'ws://{local_ip}:{self.http.port}/ws'

    async def close(self) -> None:
        async def _close():
            await asyncio.sleep(0.01)
            await self.children.close()
            await self.parent.close()

        await asyncio.sleep(0.01)
        self.io.create_task(_close())

    def capture_logs(self) -> StreamCapturing:
        """ Sets up a stream capturing context, forwarding logs to the node """
        def logger(file):
            def callback(x):
                nonlocal file
                self.io.create_task(self.parent.send_log(file, x))
            return callback

        return StreamCapturing(
            on_stdout=logger('stdout'),
            on_stderr=logger('stderr'),
            silence=True,
        )



import inspect
from .task import Task


def task(func):
    """ Wraps a task function in a Task class """

    if func.__name__[0].lower() == func.__name__[0]:
        raise NameError(
            f'Task names must start with an uppercase character, '
            f'found {func.__name__}')

    class FuncTask(Task):
        async def run(self, **inputs):
            if inspect.iscoroutinefunction(func):
                return await func(**inputs)
            else:
                return func(**inputs)

    # copy name & module
    FuncTask.__name__ = func.__name__
    FuncTask.__module__ = func.__module__
    FuncTask.__wraps__ = func

    return FuncTask



import inspect
from .type import Type
from .simple import Any
from .list import List
from .dict import Dict
from .mapping import get_type, convert_type


def convert_type_annotation(annot: object) -> Type:
    """ Converts a type annotation to a cowait type """
    if annot == inspect._empty:
        # an empty type signature defaults to the Any type.
        return Any()

    return convert_type(annot)


def get_parameter_defaults(func: callable) -> dict:
    """ Returns a dict of default parameter values function """
    sig = inspect.signature(func)
    return {
        key: parameter.default
        for key, parameter in sig.parameters.items()
        if parameter.default != inspect._empty
    }


def get_return_type(func: callable) -> Type:
    """ Gets the return type of a function. """
    sig = inspect.signature(func)
    return convert_type_annotation(sig.return_annotation)


def get_parameter_types(func: callable) -> Type:
    """
    Gets the parameter types for a function.
    Returns a Dict, mapping parameter names to cowait types.
    """
    sig = inspect.signature(func)
    return Dict({
        key: convert_type_annotation(parameter.annotation)
        for key, parameter in sig.parameters.items()
        if parameter.kind == inspect._POSITIONAL_OR_KEYWORD
    })


def type_from_description(desc: any) -> Type:
    if isinstance(desc, dict):
        return Dict({
            key: type_from_description(typedesc)
            for key, typedesc in desc.items()
        })
    elif isinstance(desc, list):
        if len(desc) == 0:
            return List()
        elif len(desc) == 1:
            list_type = type_from_description(desc[0])
            return List(list_type)
        else:
            raise TypeError('List typedef should contain a single item')
    elif isinstance(desc, str):
        # it should be a type name!
        return get_type(desc)()
    else:
        raise TypeError('Invalid type description')


def typed_arguments(func: callable, args: dict) -> dict:
    """
    Checks argument types against function signature and deserializes using cowait types.
    """
    # validate args
    parameter_types = get_parameter_types(func)
    parameter_types.validate(args, 'Arguments')

    # deeserialize args
    return parameter_types.deserialize(args)


def typed_return(func: callable, result: object) -> object:
    """
    Type checks and serializes a return value using cowait types.
    """
    result_type = get_return_type(func)
    if isinstance(result_type, Any):
        result_type = guess_type(result)

    # serialize & validate
    data = result_type.serialize(result)
    result_type.validate(data, 'Return')
    return data, result_type


async def typed_call(func: callable, args: dict) -> object:
    """ Performs a type-checked function call. """
    args = {
        **get_parameter_defaults(func),
        **args,
    }
    args = typed_arguments(func, args)
    result = func(**args)
    return typed_return(func, result)


async def typed_async_call(func: callable, args: dict) -> object:
    """ Performs a type-checked asynchronous function call. """
    args = {
        **get_parameter_defaults(func),
        **args,
    }
    targs = typed_arguments(func, args)
    result = await func(**targs)
    return typed_return(func, result)


def guess_type(obj: object) -> Type:
    """
    Attempts to guess the cowait type of an object by checking if
    its type has a registered cowait type mapping.

    Raises a TypeError if obj is not a valid cowait type.
    """
    return convert_type(type(obj))


def validate_type(obj: object, shape: Type = None) -> bool:
    """
    Validates the type of an object. If no type is provided, it will
    be inferred from the value. Returns true if the object is a valid
    cowait type.
    """
    if shape is None:
        shape = guess_type(obj)

    try:
        shape.validate(obj)
        return True
    except ValueError:
        return False


def serialize(obj: object, shape: Type = None) -> object:
    """ Best-effort serialization of an object """
    if shape is None:
        shape = guess_type(obj)

    return shape.serialize(obj)


def deserialize(obj: any, typedesc: any) -> object:
    type = type_from_description(typedesc)
    return type.deserialize(obj)



import aiohttp
from aiohttp import web
from cowait.utils import EventEmitter
from .conn import Conn


class Server(EventEmitter):
    def __init__(self, node):
        super().__init__()
        self.conns = []
        node.http.add_get('/ws', self.handle_client)

    async def handle_client(self, request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        conn = Conn(ws, request.remote)
        await self.emit(type='__connect', conn=conn)
        self.conns.append(conn)

        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    event = msg.json()

                    if conn.rpc.intercept_event(**event):
                        continue

                    await self.emit(**event, conn=conn)

        except Exception as e:
            await self.emit(type='__error', conn=conn, error=str(type(e)))

        finally:
            # disconnected
            self.conns.remove(conn)
            await self.emit(type='__close', conn=conn)

    async def send(self, msg: dict) -> None:
        for ws in self.conns:
            await ws.send_json(msg)

    async def close(self):
        for conn in self.conns:
            await conn.close()



import aiohttp
from .rpc_client import RpcClient


class Conn(object):
    def __init__(self, ws, remote):
        self.ws = ws
        self.id = None
        self.rpc = RpcClient(ws)
        self.remote = remote

    async def send(self, msg: dict) -> None:
        await self.ws.send_json(msg)

    async def close(self, message=''):
        self.rpc.cancel_all()
        await self.ws.close(
            code=aiohttp.WSCloseCode.GOING_AWAY,
            message=message
        )



import pytest
from .loader import load_task_class
from cowait.tasks import Task, TaskNotFoundError
from cowait.tasks.shell import ShellTask


class TaskOne(Task):
    pass


class TaskTwo(Task):
    pass


def test_illegal_class_name():
    with pytest.raises(TaskNotFoundError, match=r'.*Illegal.*'):
        load_task_class('cowait.')

    with pytest.raises(TaskNotFoundError, match=r'.*Illegal.*'):
        load_task_class('.cowait')


def test_load_task_class():
    loaded = load_task_class('cowait.tasks.shell')
    assert loaded == ShellTask


def test_task_module_not_found():
    with pytest.raises(TaskNotFoundError):
        load_task_class('cowait.worker.unknown_module')


def test_no_defined_tasks():
    with pytest.raises(TaskNotFoundError, match=r'.*not contain any tasks.*'):
        load_task_class('cowait.worker.loader')


def test_multiple_defined_tasks():
    with pytest.raises(TaskNotFoundError, match=r'.*multiple*.'):
        load_task_class('cowait.worker.test_loader')



from aiohttp import web
from aiohttp_middlewares import cors_middleware
from .auth_middleware import AuthMiddleware


class HttpServer():
    def __init__(self, port: int = 80):
        self.port = port
        self.auth = AuthMiddleware()

        # create http app
        self.app = web.Application(
            middlewares=[
                self.auth.middleware,
                cors_middleware(allow_all=True)
            ],
        )

        # route shortcuts
        self.add_routes = self.app.router.add_routes
        self.add_route = self.app.router.add_route
        self.add_post = self.app.router.add_post
        self.add_get = self.app.router.add_get

    async def serve(self):
        runner = web.AppRunner(self.app, handle_signals=False)
        await runner.setup()
        site = web.TCPSite(runner, host='*', port=self.port)
        await site.start()



import setuptools

VERSION = "0.3.0"

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cowait",
    version=VERSION,
    author="Backtick Technologies",
    description="Cowait is a framework for creating " +
                "containerized workflows with asynchronous Python.",

    long_description=long_description,
    long_description_content_type="text/markdown",

    url="https://github.com/backtick-se/cowait",

    packages=setuptools.find_packages(),
    scripts=['bin/cowait'],

    classifiers=[],

    python_requires='>=3.6',

    install_requires=[
        'click',
        'docker',
        'requests',
        'pyyaml',
        'dask',
        'distributed',
        'kubernetes',
        'pyyaml',
        'marshmallow',
        'aiohttp',
        'aiohttp-middlewares',
        'pytest',
        'pytest-sugar',
        'pytest-cov',
        'alt-pytest-asyncio',
        'nest-asyncio',
        'numpy',
        'sty',
        'jupyterlab',
        'dill',
        'python-dotenv',
        'fsspec',
        's3fs',
    ],
)



import string
import random


def uuid(length=8, lower=True):
    """sebbe-approved UUID"""
    # risk of collision
    # mixed case: 8 characters -> 1 in 54 trillion
    # lower case: 8 characters -> 1 in 208 billion

    letters = string.ascii_letters
    if lower:
        letters = letters.lower()
    return ''.join(random.choice(letters) for i in range(length))



import pytest
from .const import MAX_ENV_LENGTH
from .cluster import ClusterProvider
from .errors import ProviderError
from cowait.tasks import TaskDefinition
from cowait.utils import uuid


def test_max_env_length():
    """ Passing too large inputs should raise a ProviderError """
    random_data = uuid(2 * MAX_ENV_LENGTH, lower=False)

    with pytest.raises(ProviderError):
        cp = ClusterProvider('test')
        cp.create_env(TaskDefinition(
            'test-task',
            image='imaginary-image',
            inputs={
                'ohshit': random_data,
            },
        ))



from cowait.network import Conn
from cowait.utils import EventEmitter


class Subscriptions(EventEmitter):
    """ Websocket subscriber component """

    def __init__(self, task):
        super().__init__()
        self.subscribers = []
        self.__len__ = self.subscribers.__len__
        self.__str__ = self.subscribers.__str__
        self.__iter__ = self.subscribers.__iter__
        self.__getitem__ = self.subscribers.__getitem__
        self.__contains__ = self.subscribers.__contains__

        task.node.children.on('subscribe', self.subscribe)
        task.node.children.on('unsubscribe', self.unsubscribe)
        task.node.children.on('__close', self.unsubscribe)
        task.node.children.on('*', self.forward)

    async def forward(self, conn: Conn, **msg):
        if conn in self.subscribers:
            return

        if msg.get('type', '__')[:2] == '__':
            return

        for subscriber in self.subscribers:
            await subscriber.send(msg)

    async def subscribe(self, conn: Conn, **msg):
        print(f'~~ add subscriber {conn.remote}')
        self.subscribers.append(conn)
        await self.emit(
            type='subscribe',
            conn=conn,
        )

    async def unsubscribe(self, conn: Conn, **msg):
        if conn in self.subscribers:
            print(f'~~ drop subscriber {conn.remote}')
            self.subscribers.remove(conn)
            await self.emit(
                type='unsubscribe',
                conn=conn,
            )
        else:
            print(f'~~ drop task {conn.remote}')



from __future__ import annotations
from marshmallow import fields, post_load
from .status import WAIT
from .definition import TaskDefinition, TaskDefinitionSchema


class TaskInstance(TaskDefinition):
    def __init__(
        self,
        status: str = WAIT,
        error: str = None,
        result: any = None,
        log: str = '',
        **data,
    ):
        super().__init__(**data)
        self.status = status
        self.error = error
        self.result = result
        self.log = log

    def serialize(self) -> dict:
        """ Serialize task definition to a dict """
        return TaskInstanceSchema().dump(self)

    @staticmethod
    def deserialize(instance: dict) -> TaskInstance:
        """ Deserialize task definition from a dict """
        return TaskInstanceSchema().load(instance)


class TaskInstanceSchema(TaskDefinitionSchema):
    status = fields.Str(missing=WAIT)
    error = fields.Str(allow_none=True)
    result = fields.Raw(allow_none=True)
    log = fields.Str(allow_none=True)

    @post_load
    def make_taskinstance(self, data: dict, **kwargs) -> TaskInstance:
        return TaskInstance(**data)
